# -*- coding: utf-8 -*-
import sys
from functools import partial
import logging as python_logging
from http.client import HTTPConnection
from pprint import pprint

from eliot import Message, write_traceback, add_destinations
from eliot.stdlib import EliotHandler

from live_client.connection.rest_input import async_event_sender

__all__ = ["debug", "info", "warn", "error", "exception", "log_to_live", "log_to_stdout"]

LOG_LEVELS = ["DEBUG", "INFO", "WARN", "ERROR", "EXCEPTION"]


default_level = "INFO"
log_level = default_level


def log_message(message, severity=None):
    if level_is_logged(severity):
        return Message.log(message_type=severity, message=message)


def exception(message):
    write_traceback(exc_info=sys.exc_info())


debug = partial(log_message, severity="debug")
info = partial(log_message, severity="info")
warn = partial(log_message, severity="warn")
error = partial(log_message, severity="error")


def log_to_live(message, log_function=None, event_type=None, min_level=default_level):
    message_severity = message.get("message_type", min_level)
    if level_is_logged(message_severity, min_level=min_level):
        message.update(__type=event_type)
        log_function(message)


def log_to_stdout(message, min_level=default_level):
    message_severity = message.get("message_type", min_level)
    if level_is_logged(message_severity, min_level=min_level):
        pprint(message)


def level_is_logged(message_severity, min_level=None):
    severity = message_severity.upper()
    if severity in LOG_LEVELS:
        message_severity_idx = LOG_LEVELS.index(severity)

        if (min_level is None) or (min_level.upper() not in LOG_LEVELS):
            min_level = log_level
        else:
            min_level = min_level.upper()

        min_level_idx = LOG_LEVELS.index(min_level.upper())

        result = message_severity_idx >= min_level_idx
    else:
        result = True

    return result


def setup_live_logging(logging_settings, live_settings):
    event_type = logging_settings.get("event_type", "dda_log")
    level = logging_settings.get("level", default_level)

    username = live_settings.get("username")
    password = live_settings.get("password")
    logging_url = f"{live_settings['url']}{live_settings['rest_input']}"

    global log_level
    log_level = level
    log_function = async_event_sender(live_settings)

    if event_type and logging_url and username and password:
        add_destinations(
            partial(log_to_live, log_function=log_function, event_type=event_type, min_level=level)
        )


def setup_python_logging(logging_settings):
    level = logging_settings.get("level", default_level)

    log_level = getattr(python_logging, level.upper())

    if log_level is python_logging.DEBUG:
        HTTPConnection.debuglevel = 1

    python_logging.basicConfig()
    python_logging.getLogger().setLevel(log_level)
    python_logging.getLogger().addHandler(EliotHandler())

    requests_log = python_logging.getLogger("urllib3")
    requests_log.addHandler(EliotHandler())
    requests_log.setLevel(log_level)
    requests_log.propagate = False

    cometd_log = python_logging.getLogger("aiocometd")
    cometd_log.addHandler(EliotHandler())
    cometd_log.setLevel(log_level)
    cometd_log.propagate = False



# -*- coding: utf-8 -*-
import socket
from contextlib import contextmanager

from requests.exceptions import Timeout

from live_client.utils import logging

__all__ = ["ensure_timeout"]


@contextmanager
def ensure_timeout(timeout):
    default_timeout = socket.getdefaulttimeout()
    if isinstance(timeout, (list, tuple)):
        socket_timeout = timeout[0]
    else:
        socket_timeout = timeout

    if socket_timeout is not None:
        socket.setdefaulttimeout(socket_timeout)

    logging.debug(f"Socket timeout is now {socket_timeout}")

    try:
        yield
    finally:
        socket.setdefaulttimeout(default_timeout)
        logging.debug(f"Socket timeout back to default ({default_timeout})")


@contextmanager
def retry_on_failure(timeout=3.05, max_retries=0):
    request_finished = False
    retries = 0

    while request_finished is False:
        try:
            with ensure_timeout(timeout):
                yield
        except (socket.timeout, Timeout):
            if max_retries and (retries < max_retries):
                logging.info(f"Operation timed out, retrying ({retries}/{max_retries})")
                retries += 1
                continue
            else:
                logging.error(f"Operation timed out")
        finally:
            request_finished = True



# -*- coding: utf-8 -*-

EVENT_TYPE_EVENT = "event"
EVENT_TYPE_SPAN = "span"
EVENT_TYPE_STOP = "stop"
EVENT_TYPE_DESTROY = "destroy"

UOM_KEY = "uom"
VALUE_KEY = "value"
TIMESTAMP_KEY = "timestamp"

DEFAULT_ANNOTATION_DURATION = 30000



# -*- coding: utf-8 -*-
from multiprocessing import Queue

from live_client import query
from live_client.utils import logging, http

__all__ = ["list_assets", "fetch_asset_settings", "watch_asset_settings"]


ALL_ASSET_TYPES = ["rig", "crew", "pump"]


def list_assets(settings, asset_type=None):
    live_settings = settings["live"]
    url = live_settings["url"]
    data = []

    if asset_type in ALL_ASSET_TYPES:
        chosen_asset_types = [asset_type]

    elif asset_type is None:
        chosen_asset_types = ALL_ASSET_TYPES

    else:
        chosen_asset_types = []

    for atype in chosen_asset_types:
        asset_url = f"{url}/services/plugin-liverig/assets/{atype}"
        try:
            response_data = http.request_with_timeout(asset_url, settings)
            if response_data is not None:
                for asset in response_data:
                    asset["asset_type"] = atype
                    data.append(asset)
        except Exception:
            logging.exception(f"Error fetching asset list for {atype}")

    return data


def fetch_asset_settings(settings, asset_id, asset_type="rig"):
    live_settings = settings["live"]
    url = live_settings["url"]
    asset_url = f"{url}/services/plugin-liverig/assets/{asset_type}/{asset_id}/normalizer"
    return http.request_with_timeout(asset_url, settings)


def watch_asset_settings(settings, asset_id, asset_type="rig"):
    initial_config = fetch_asset_settings(settings, asset_id, asset_type=asset_type)

    asset_query_template = "__asset event:(normalizer|delete) type:{} id:{}"
    asset_query = asset_query_template.format(asset_type, asset_id)
    results_process, results_queue = query.run(settings, asset_query, realtime=True)

    results_queue.put({"config": initial_config})
    return _settings_update_handler(results_queue)


def _settings_update_handler(events_queue):
    settings_queue = Queue()

    while True:
        event = events_queue.get()
        settings_queue.put(event.get("config", {}))

    return settings_queue



# -*- coding: utf-8 -*-
from functools import partial

from . import rest_input, tcp_input

__all__ = ["send_event"]


def has_required_keys(settings, key_list):
    return all([item in settings for item in key_list])


def send_event(event, live_settings=None):
    if live_settings is None:
        live_settings = {}

    if has_required_keys(live_settings, rest_input.REQUIRED_PARAMETERS):
        sender_function = rest_input.send_event
    elif has_required_keys(live_settings, tcp_input.REQUIRED_PARAMETERS):
        sender_function = tcp_input.send_event
    else:
        raise ValueError(
            "Invalid settings. The keys '{}' or '{}' must be defined".format(
                rest_input.REQUIRED_PARAMETERS, tcp_input.REQUIRED_PARAMETERS
            )
        )

    sender_function(event, live_settings=live_settings)


def build_sender_function(live_settings):
    return partial(send_event, live_settings=live_settings)



# -*- coding: utf-8 -*-
from functools import partial
from requests.exceptions import RequestException

from eliot import start_action

from live_client.connection.rest_input import build_session
from live_client.utils.network import retry_on_failure
from live_client.utils import logging

__all__ = ["make_request", "request_with_timeout"]


def make_request(url, settings, timeout=None, max_retries=0):
    live_settings = settings["live"]
    verify_ssl = live_settings.get("verify_ssl", True)

    if "session" not in live_settings:
        live_settings.update(session=build_session(live_settings))

    session = live_settings["session"]

    with start_action(action_type="make request", url=url):
        with retry_on_failure(timeout, max_retries=max_retries):
            try:
                response = session.get(url, verify=verify_ssl)
                response.raise_for_status()
                result = response.json()

            except RequestException as e:
                logging.exception(f"Error during request for {url}, {e}<{type(e)}>")
                result = None

    return result


request_with_timeout = partial(make_request, timeout=(3.05, 5), max_retries=5)



# -*- coding: utf-8 -*-
import requests
from multiprocessing import Process, Queue

from requests.exceptions import RequestException
from setproctitle import setproctitle
from eliot import start_action

from live_client.utils import logging
from live_client.utils.network import retry_on_failure

__all__ = ["send_event"]

REQUIRED_PARAMETERS = ["url", "username", "password", "rest_input"]


def build_session(live_settings):
    username = live_settings["username"]
    password = live_settings["password"]

    session = requests.Session()
    session.auth = (username, password)

    return session


def send_event(event, live_settings=None):
    if live_settings is None:
        live_settings = {}

    if "session" not in live_settings:
        live_settings.update(session=build_session(live_settings))

    session = live_settings["session"]
    verify_ssl = live_settings.get("verify_ssl", True)
    url = f"{live_settings['url']}{live_settings['rest_input']}"

    if not event:
        return

    try:
        with retry_on_failure(3.05, max_retries=5):
            response = session.post(url, json=event, verify=verify_ssl)
            response.raise_for_status()
    except RequestException as e:
        logging.exception("ERROR: Cannot send event, {}<{}>".format(e, type(e)))
        logging.exception("Event data: {}".format(event))
        raise


def async_send(queue, live_settings):
    with start_action(action_type="async_logger"):
        logging.info("Remote logger process started")
        setproctitle("DDA: Remote logger")

        live_settings.update(session=build_session(live_settings))
        while True:
            event = queue.get()
            send_event(event, live_settings)


def async_event_sender(live_settings):
    events_queue = Queue()
    process = Process(target=async_send, args=(events_queue, live_settings))
    process.start()

    return lambda event: events_queue.put(event)



# -*- coding: utf-8 -*-
from live_client.resources.base import fetch_resource

__all__ = ["fetch_role_info", "fetch_roles_info"]


def fetch_role_info(role_id, settings):
    return fetch_roles_info([role_id], settings)


def fetch_roles_info(role_ids, settings):
    # Fetch all roles. Larger payload instead of more requests
    response = fetch_resource("/rest/role", settings)
    all_roles = response.get("data", [])
    return [item for item in all_roles if item["id"] in role_ids]



from . import query
from .events import messenger
from .utils.timestamp import get_timestamp


DEFAULT_REQUEST_TIMEOUT = (3.05, 5)
DEFAULT_MAX_RETRIES = 5


def merge_extra_settings(func):
    def ret(*args, **kwargs):
        extra_settings = kwargs.pop("extra_settings", {})
        if extra_settings:
            settings = kwargs.get("settings", {})
            settings.update(extra_settings)
            kwargs["settings"] = settings
        return func(*args, **kwargs)

    return ret


@merge_extra_settings
def send_message(*args, **kwargs):
    return messenger.send_message(*args, **kwargs)


class LiveClient:
    def __init__(self, settings, room_id):
        self.settings = settings
        self.room_id = room_id

    def run_query(self, query_str, realtime, span=None):
        return query.run(
            query_str,
            self.settings,
            realtime=realtime,
            span=span,
            timeout=DEFAULT_REQUEST_TIMEOUT,
            max_retries=DEFAULT_MAX_RETRIES,
        )

    def send_message(self, message):
        return send_message(
            message, get_timestamp(), settings=self.settings, room={"id": self.room_id}
        )



# -*- coding: utf-8 -*-
from live_client.resources.base import fetch_resource
from .role import fetch_roles_info

__all__ = ["fetch_team_info", "fetch_teams_info"]


def fetch_team_info(team_id, settings):
    return fetch_teams_info([team_id], settings)


def fetch_teams_info(team_ids, settings):
    # Fetch all teams. Larger payload instead of more requests
    response = fetch_resource("/rest/team", settings)
    all_teams = response.get("data", [])
    selected_teams = [item for item in all_teams if item["id"] in team_ids]

    # Fetch the data for roles related to the teams
    role_ids = []
    for item in selected_teams:
        role_ids.extend(role["id"] for role in item.get("roles", []))

    selected_roles = fetch_roles_info(set(role_ids), settings)

    # Add the roles for each team
    roles_by_id = dict((item["id"], item) for item in selected_roles)

    for team in selected_teams:
        team.update(roles=[roles_by_id.get(item["id"]) for item in team.get("roles", [])])

    return selected_teams



# -*- coding: utf-8 -*-
import sys
import socket
import json

from live_client.utils import logging

__all__ = ["send_event"]

REQUIRED_PARAMETERS = ["ip", "port"]


def send_event(event, live_settings=None):
    if live_settings is None:
        live_settings = {}

    ip = live_settings["ip"]
    port = live_settings["port"]

    if not event:
        return

    message = "{}\n".format(json.dumps(event))
    python_version = sys.version_info.major
    if python_version == 3:
        message = bytes(message, "utf-8")

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((ip, port))
        sock.sendall(message)
    except socket.error:
        logging.exception("ERROR: Cannot send event, server unavailable")
        logging.exception("Event data: {}".format(message))
    finally:
        sock.close()



# -*- coding: utf-8 -*-
from .base import fetch_resource

__all__ = ["list_plugins"]


def list_plugins(settings, include_disabled=False):
    plugins_info = fetch_resource("/rest/plugin", settings)

    if plugins_info is None:
        plugins_info = {}

    def is_enabled(plugin_data):
        return plugin_data.get("status", {}).get("status") == "VALID"

    plugins_list = [
        item for item in plugins_info.get("data", []) if include_disabled or is_enabled(item)
    ]

    return plugins_list



# -*- coding: utf-8
from .resources.access_control import permissions

REQUIREMENTS = {
    "live_client.query": {"permissions": [permissions.CONSOLE]},
    "live_client.assets": {
        "plugins": ["plugin-liverig>=2.17.0"],
        "permissions": [permissions.ASSETS],
    },
    "live_client.assets.auto_analysis": {
        "plugins": [
            "plugin-liverig-vis>=2.17.0",
            "plugin-annotations>=2.24.5",
            "plugin-restinput>=2.25.0",
        ],
        "permissions": [permissions.ASSETS],
    },
    "live_client.connection.rest_input": {"plugins": ["plugin-restinput>=2.25.0"]},
    "live_client.connection.tcp_input": {"plugins": ["plugin-tcpinput>=2.24.0"]},
    "live_client.events.annotation": {
        "plugins": ["plugin-annotations>=2.24.5", "plugin-restinput>=2.25.0"],
        "permissions": [permissions.EDIT_ANNOTATIONS],
    },
    "live_client.events.messenger": {"plugins": ["plugin-messenger>=2.24.0"]},
    "live_client.resources.dashboards": {
        "plugins": ["live-ui>=2.24.0"],
        "permissions": [permissions.DASHBOARDS],
    },
    "live_client.resources.messenger": {
        "plugins": ["plugin-messenger>=2.24.0"],
        "permissions": [permissions.USE_MESSENGER],
    },
}



# -*- coding: utf-8 -*-
import uuid
from enum import Enum

from . import raw
from live_client.connection.autodetect import build_sender_function
from live_client.utils.timestamp import get_timestamp
from live_client.utils import logging

__all__ = ["send_message", "maybe_send_message_event", "maybe_send_chat_message", "format_and_send"]


MESSAGE_TYPES = Enum("MESSAGE_TYPES", "EVENT, CHAT")
CONTROL_ACTIONS = Enum("CONTROL_ACTIONS", "ADD_USER, REMOVE_USER, JOIN, LEAVE")
EVENT_TYPES = {"message": "__message", "control": "__messenger_control"}


__all__ = [
    "join_messenger",
    "add_to_room",
    "remove_from_room",
    "send_message",
    "maybe_send_chat_message",
    "maybe_send_message_event",
]


def join_messenger(settings):
    connection_func = build_sender_function(settings["live"])

    bot_data = settings["output"]["author"]
    control_data = {
        "broadcast": True,
        "__skipstorage": True,
        "action": "user_joined_messenger",
        "user": bot_data,
    }

    event = format_event(control_data, EVENT_TYPES["control"], get_timestamp())
    connection_func(event)


def add_to_room(settings, room_id, sender):
    return update_room_users(settings, room_id, sender, action=CONTROL_ACTIONS.ADD_USER)


def remove_from_room(settings, room_id, sender):
    return update_room_users(settings, room_id, sender, action=CONTROL_ACTIONS.REMOVE_USER)


def update_room_users(settings, room_id, sender, action):
    connection_func = build_sender_function(settings["live"])
    author = settings["output"]["author"]

    update_message_payload = create_room_update_data(room_id, sender, author, action)
    update_event = format_event(update_message_payload, EVENT_TYPES["control"], get_timestamp())
    connection_func(update_event)


def create_room_update_data(room_id, sender, author, action):
    target_collection = {
        CONTROL_ACTIONS.ADD_USER: "addedOrUpdatedUsers",
        CONTROL_ACTIONS.REMOVE_USER: "removedUsers",
    }

    data = {
        "action": "room_users_updated",
        "sender": sender,
        "addedOrUpdatedUsers": [],
        "removedUsers": [],
        "room": {"id": room_id},
    }

    if action == CONTROL_ACTIONS.ADD_USER:
        author.update(isNewUser=True, isAdmin=False)

    data[target_collection[action]].append(author)

    return data


# [ECS][FIXME]: This function may choose none, only one or both functions.
#   Even when one of them or both are choosen neither may be called depending on internal rules.
#   It is way too complex and should be eliminated
#   Clients should explicitly call the version they want.
def send_message(message, **kwargs):
    timestamp = kwargs.pop("timestamp", get_timestamp())
    message_type = kwargs.pop("message_type", None)

    if (message_type is None) or (message_type == MESSAGE_TYPES.EVENT):
        maybe_send_message_event(message, timestamp, kwargs.get("settings"))

    if (message_type is None) or (message_type == MESSAGE_TYPES.CHAT):
        maybe_send_chat_message(message, **kwargs)


def maybe_send_message_event(message, timestamp, settings):
    output_settings = settings["output"]
    message_event = output_settings.get("message_event", {})
    event_type = message_event.get("event_type")
    messages_mnemonic = message_event.get("mnemonic")

    if event_type and messages_mnemonic:
        connection_func = build_sender_function(settings["live"])
        event = {"timestamp": timestamp, messages_mnemonic: {"value": message}}
        logging.debug("Sending message event '{}' for '{}'".format(event, event_type))
        raw.format_and_send(event, event_type, connection_func=connection_func)
        return True

    return False


def maybe_send_chat_message(message, settings, **kwargs):
    output_settings = settings["output"]
    author = output_settings.get("author")
    room = kwargs.get("room", output_settings.get("room"))

    shall_send_message = (room is not None) and (author is not None)

    if not shall_send_message:
        logging.warn(
            f"Cannot send message, room ({room}) and/or author ({author}) missing. Message is '{message}'"
        )
        return False

    # [ECS][FIXME]: Author should not be altered here. It'd be better to receive it configured from the client <<<<<
    author["name"] = kwargs.get("author_name") or author.get("name")
    connection_func = build_sender_function(settings["live"])
    logging.debug("Sending message '{}' from {} to {}".format(message, author, room))
    format_and_send(message, room, author, connection_func=connection_func)
    return True


def format_and_send(message, room, author, connection_func):
    event = format_message_event(message, room, author, timestamp=get_timestamp())
    logging.debug("Sending message {}".format(event))
    connection_func(event)


def format_message_event(message, room, author, timestamp):
    message_data = {"message": message, "room": room, "author": author}
    return format_event(message_data, EVENT_TYPES["message"], timestamp)


def format_event(event_data, event_type, timestamp):
    base_event = {"__type": event_type, "uid": str(uuid.uuid4()), "createdAt": timestamp}
    base_event.update(event_data)
    return base_event



import sys

sys.path.insert(1, "..")

from datetime import datetime
from live_client.utils.timestamp import get_timestamp


def no_action(*args, **kwargs):
    pass


def raises(exception_cls, f, *args, **kwargs):
    try:
        f(*args, **kwargs)
    except exception_cls:
        return True

    return False


class Collector:
    def __init__(self):
        self.buffer = []

    def collect(self, obj):
        self.buffer.append(obj)

    def __call__(self, obj):
        self.collect(obj)

    def is_empty(self):
        return len(self.buffer) == 0

    def __getitem__(self, key):
        return self.buffer[key]



# -*- coding: utf-8 -*-
import json

from live_client.utils.timestamp import get_timestamp

__all__ = ["format_and_send"]


def format_and_send(event_type, statuses, output_settings):
    timestamp = get_timestamp()
    event = format_event(timestamp, event_type, statuses)
    print("-----------------------------------------------------")
    print(json.dumps(event))


def format_event(timestamp, event_type, statuses):
    event_data = statuses.copy()
    event_data["__type"] = event_type
    event_data["liverig__index__timestamp"] = timestamp
    return json.dumps(event_data)



# -*- coding: utf-8 -*-
import uuid

from live_client.connection import autodetect
from live_client.utils.timestamp import get_timestamp
from live_client.utils import logging
from .constants import DEFAULT_ANNOTATION_DURATION

__all__ = ["create", "format_and_send"]


def create(annotation_data, settings=None, room=None):
    output_settings = settings["output"]
    connection_func = autodetect.build_sender_function(settings["live"])

    if room is None:
        room = output_settings["room"]

    output_settings.update(
        room=room, author=output_settings["author"], dashboard=output_settings.get("dashboard", {})
    )
    format_and_send(annotation_data, output_settings, connection_func=connection_func)


def format_and_send(annotation_data, settings, connection_func=None):
    timestamp = get_timestamp()
    event = format_event(timestamp, annotation_data, settings)

    logging.debug("Creating annotation {}".format(event))
    connection_func(event)


def format_event(timestamp, annotation_data, settings):
    author_data = settings["author"]
    room_data = settings["room"]
    dashboard_data = settings["dashboard"]

    message_event = annotation_data.copy()
    message_event.update(
        __type="__annotations",
        __src=message_event.get("__src", "live_agent"),
        uid=message_event.get("uid", str(uuid.uuid4())),
        createdAt=int(message_event.get("createdAt", timestamp)),
        author=author_data.get("name"),
        room=room_data,
        dashboardId=dashboard_data.get("id"),
        dashboard=dashboard_data.get("name"),
        searchable=True,
    )

    begin = int(message_event.get("begin", timestamp))

    end = message_event.pop("end", None)
    if (end is None) or end < begin:
        end = begin + DEFAULT_ANNOTATION_DURATION

    message_event.update(begin=begin)
    message_event.update(end=int(end))

    def has_invalid_value(key):
        return message_event.get(key, -1) in (0, None)

    if any(map(has_invalid_value, ("begin", "end", "createdAt"))):
        logging.warn("Invalid annotation: {}".format(message_event))
        return

    return message_event



from unittest import mock

from testbase import *

from live_client.events import messenger
from predicates import *
from mocks import *


def gen_settings(**kwargs):
    default_settings = {
        "output": {"author": {"id": 1, "name": "__default_name__"}, "room": "__room__"},
        "live": {},
    }
    default_settings.update(kwargs)
    return default_settings


class TestJoinMessenger:
    def test_correct_action_sent(self):
        settings = gen_settings()

        connection_mock = mock.Mock()
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", connection_mock
        ):
            messenger.join_messenger(settings)
            event = connection_mock.call_args[0][0]
            assert event["action"] == "user_joined_messenger"

    def test_correct_author_sent(self):
        settings = gen_settings()
        author = settings["output"]["author"]

        connection_mock = mock.Mock()
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", connection_mock
        ):
            messenger.join_messenger(settings)
            event = connection_mock.call_args[0][0]
            assert event["user"] == author


class TestAddToRoom:
    def test_correct_action_sent(self):
        settings = gen_settings()
        connection_mock = mock.Mock()

        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", connection_mock
        ):
            messenger.add_to_room(settings, 0, "_")
            event = connection_mock.call_args[0][0]
            assert event["action"] == "room_users_updated"

    def test_user_is_added(self):
        user = {"id": 2, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        chat_mock = ChatMock()
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.add_to_room(settings, room_id, "Tester")

            assert chat_mock.room != {}
            assert chat_mock.room["users"][user["id"]]["name"] == user["name"]

    def test_user_is_not_removed(self):
        user = {"id": 1, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        existing_user = {"id": 2, "name": "__should_remain__"}
        chat_mock = ChatMock()
        chat_mock.add_user(existing_user)
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.add_to_room(settings, room_id, "Tester")
            assert (
                not is_empty(chat_mock.room["users"])
                and chat_mock.room["users"][2]["name"] == existing_user["name"]
            )


class TestRemoveFromRoom:
    def test_correct_action_sent(self):
        settings = gen_settings()
        connection_mock = mock.Mock()

        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", connection_mock
        ):
            messenger.remove_from_room(settings, 0, "_")
            event = connection_mock.call_args[0][0]
            assert event["action"] == "room_users_updated"

    def test_user_is_removed(self):
        user = {"id": 2, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        chat_mock = ChatMock()
        chat_mock.room["users"][2] = user
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.remove_from_room(settings, room_id, "Tester")
            assert chat_mock.room["users"] == {}

    def test_user_is_not_removed(self):
        user = {"id": 1, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        existing_user = {"id": 2, "name": "__should_remain__"}
        chat_mock = ChatMock()
        chat_mock.add_user(existing_user)
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.remove_from_room(settings, room_id, "Tester")
            assert (
                bool(chat_mock.room["users"])
                and chat_mock.room["users"][2]["name"] == existing_user["name"]
            )


class TestUpdateRoomUsers:
    def test_user_is_added(self):
        user = {"id": 2, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        chat_mock = ChatMock()
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.update_room_users(
                settings,
                room_id=room_id,
                sender="Tester",
                action=messenger.CONTROL_ACTIONS.ADD_USER,
            )

            assert len(chat_mock.room) > 0
            assert chat_mock.room["users"][user["id"]]["name"] == user["name"]

    def test_user_is_removed(self):
        user = {"id": 2, "name": "__local_test__"}
        settings = gen_settings()
        settings["output"]["author"] = user

        room_id = 1
        chat_mock = ChatMock()
        chat_mock.room["users"][2] = user
        with patch_with_factory(
            "live_client.events.messenger.build_sender_function", chat_mock.update_room
        ):
            messenger.update_room_users(
                settings,
                room_id=room_id,
                sender="Tester",
                action=messenger.CONTROL_ACTIONS.REMOVE_USER,
            )

            assert chat_mock.room["users"] == {}


class TestSendMessage:
    @mock.patch("live_client.events.messenger.maybe_send_message_event")
    @mock.patch("live_client.events.messenger.maybe_send_chat_message")
    def test_correct_implementation_called(self, chat_mock, event_mock):
        def reset_mocks():
            nonlocal event_mock, chat_mock
            event_mock.reset_mock()
            chat_mock.reset_mock()

        # None shall be called:
        messenger.send_message("_", message_type="")
        assert not event_mock.called and not chat_mock.called
        reset_mocks()

        # send event shall be called:
        messenger.send_message("_", message_type=messenger.MESSAGE_TYPES.EVENT)
        assert event_mock.called and not chat_mock.called
        reset_mocks()

        # send chat shall be called:
        messenger.send_message("_", message_type=messenger.MESSAGE_TYPES.CHAT)
        assert not event_mock.called and chat_mock.called
        reset_mocks()

        # both shall be called:
        messenger.send_message("_", message_type=None)
        assert event_mock.called and chat_mock.called


class TestMaybeSendMessageEvent:
    # If settings["output"] is not a dictionary -> Throw exception
    def test_invalid_settings_throws(self):
        settings = {}
        assert raises(KeyError, messenger.maybe_send_message_event, "_", 0, settings)

    # Message shall be sent if:
    #   1 - event_type is not None and
    #   2 - messages_mnemonic is not None
    # Event must follow the template: event = {"timestamp": timestamp, messages_mnemonic: {"value": message}}
    @mock.patch("live_client.utils.logging.debug", no_action)
    def test_message_should_be_sent(self):
        messages_mnemonic = "__mnemonic__"

        settings = gen_settings()
        settings["output"]["message_event"] = {
            "event_type": "__event_type__",
            "mnemonic": messages_mnemonic,
        }
        collector = Collector()
        with patch_with_factory("live_client.events.messenger.build_sender_function", collector):
            message = "_"
            timestamp = get_timestamp()
            message_sent = messenger.maybe_send_message_event(message, timestamp, settings)
            assert message_sent

            sent_event = collector[0]
            assert sent_event.get("liverig__index__timestamp") == timestamp
            assert sent_event.get(messages_mnemonic) is not None
            assert sent_event.get(messages_mnemonic).get("value") == message

    def test_message_not_sent_if_no_event_type(self):
        settings = gen_settings()
        settings["output"]["message_event"] = {"mnemonic": "_"}
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message = "_"
            timestamp = get_timestamp()
            message_sent = messenger.maybe_send_message_event(message, timestamp, settings)
            assert not message_sent
            assert collector.is_empty()

    def test_message_not_sent_if_no_messages_mnemonic(self):
        settings = gen_settings()
        settings["output"]["message_event"] = {"event_type": "__event_type__"}
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message = "_"
            timestamp = get_timestamp()
            message_sent = messenger.maybe_send_message_event(message, timestamp, settings)
            assert not message_sent
            assert collector.is_empty()


class TestMaybeSendChatMessage:
    # If settings["output"] is not a dictionary -> Throw exception
    @mock.patch("live_client.connection.autodetect.build_sender_function", lambda _: no_action)
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_invalid_settings_throws(self):
        # Must fail if "output" attribute missing
        settings = {}
        assert raises(KeyError, messenger.maybe_send_chat_message, "_", settings)

    # Message shall be sent if:
    #   (1) author is configured in settings and
    #   (2) room is configured either in kwargs or in settings
    @mock.patch("live_client.events.messenger.build_sender_function", lambda _: Collector())
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_message_should_be_sent(self):
        settings = {"output": {"author": {}, "room": "__room__"}, "live": {}}

        message_sent = messenger.maybe_send_chat_message("_", settings)
        assert message_sent

    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_message_not_sent_if_no_room(self):
        settings = {"output": {"author": {}}, "live": {}}
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message_sent_1 = messenger.maybe_send_chat_message("_", settings)
            assert not message_sent_1 and collector.is_empty()

        settings["output"]["room"] = None
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message_sent_2 = messenger.maybe_send_chat_message("_", settings)
            assert not message_sent_2 and collector.is_empty()

    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_message_not_sent_if_no_author(self):
        settings = {"output": {"room": "__room__"}, "live": {}}
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message_sent_1 = messenger.maybe_send_chat_message("_", settings)
            assert not message_sent_1 and collector.is_empty()

        settings["output"]["author"] = None
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            message_sent_2 = messenger.maybe_send_chat_message("_", settings)
            assert not message_sent_2 and collector.is_empty()

    # If kwargs contains "author_name" then "author" name shall be updated.
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_author_name_updates_author(self):
        settings = gen_settings()
        new_author_name = "__new_author__"
        assert settings["output"]["author"]["name"] != new_author_name
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            messenger.maybe_send_chat_message("_", settings, author_name=new_author_name)
        assert collector.buffer[0]["author"]["name"] == new_author_name

    # If message can be sent:
    #   log success message (debug)
    #   call sender function
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_sender_called_on_success(self):
        settings = gen_settings()
        collector = Collector()
        with mock.patch("live_client.events.messenger.build_sender_function", lambda _: collector):
            messenger.maybe_send_chat_message("_", settings)
            assert not collector.is_empty()

    #   return True
    @mock.patch("live_client.events.messenger.build_sender_function", lambda _: no_action)
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_returns_true_on_success(self):
        settings = gen_settings()
        ret = messenger.maybe_send_chat_message("_", settings)
        assert ret is True

    # Else:
    #   log error message (warn)
    #   return False
    @mock.patch("live_client.events.messenger.build_sender_function", lambda _: no_action)
    @mock.patch("live_client.utils.logging.debug", no_action)
    @mock.patch("live_client.utils.logging.warn", no_action)
    def test_returns_true_on_success(self):
        settings = gen_settings(output={})
        ret = messenger.maybe_send_chat_message("_", settings)
        assert ret is False


class TestFormatAndSend:
    def mock_connection_func(self, event):
        self.event_data = event

    @mock.patch("live_client.utils.logging.debug")
    def test_message_logged(self, debug_mock):
        message = "_"
        room = "_"
        author = {}
        connection_func = self.mock_connection_func

        messenger.format_and_send(message, room, author, connection_func)
        debug_mock.assert_called_with("Sending message {}".format(self.event_data))

    def test_connection_func_called(self):
        message = "__message__"
        room = "__room__"
        author = {}
        messenger.format_and_send(message, room, author, self.mock_connection_func)
        assert (
            self.event_data["message"] == message
            and self.event_data["room"] == room
            and self.event_data["author"] == author
            and self.event_data.get("createdAt") is not None
            and self.event_data.get("uid") is not None
            and self.event_data.get("__type") is not None
        )


class TestFormatMessageEvent:
    def test_event_configured(self):
        message = "__message__"
        room = "__room__"
        author = {}
        timestamp = get_timestamp()

        base_message = {"message": message, "room": room, "author": author}
        event = messenger.format_message_event(message, room, author, timestamp)

        assert dict_contains(event, base_message) and event.get("createdAt") == timestamp


class TestFormatEvent:
    # [ECS]: Do we really need to create a new dict? <<<<<
    def test_formatted_event_is_not_the_original(self):
        event = {}
        formatted_event = messenger.format_event(event, "__type__", 0)
        assert event is not formatted_event

    def test_base_event_configured(self):
        required_keys = ["uid", "__type", "createdAt"]
        event = messenger.format_event({}, "__type__", 0)
        assert dict_has_valid_items(event, required_keys)

    def test_formatted_event_has_event_type(self):
        event_type = "__event_type__"
        event = messenger.format_event({}, event_type, 0)
        assert event.get("__type") == event_type

    def test_formatted_event_has_timestamp(self):
        timestamp = get_timestamp()
        event = messenger.format_event({}, "_", timestamp)
        assert event.get("createdAt") == timestamp



from unittest import mock

from testbase import *

from live_client.events import raw


DEFAULT_EVENT = {}
A_TIMESTAMP = 1582927835001
DEFAULT_TIMESTAMP = 0


class TestCreateEvent:
    @mock.patch("live_client.connection.autodetect.build_sender_function", lambda _: no_action)
    @mock.patch("live_client.utils.logging.debug")
    def test_message_logged(self, debug_mock):
        event_type = "__event_type__"
        event_data = {}
        settings = {"live": "__live__"}
        raw.create(event_type, event_data, settings)
        debug_mock.assert_called_with(f'Creating raw event of type "{event_type}": {event_data}')

    @mock.patch("live_client.utils.logging.debug", no_action)
    def test_event_sent(self):
        buffer = []

        def collect(event):
            nonlocal buffer
            buffer.append(event)

        with mock.patch(
            "live_client.connection.autodetect.build_sender_function", lambda _: collect
        ):
            event_type = "__event_type__"
            event_data = {"data": {}}
            settings = {"live": "__live__"}

            raw.create(event_type, event_data, settings)

            assert len(buffer) > 0
            assert buffer[0]["data"] is event_data["data"]


class TestFormatEvent:
    def test_formatted_event_is_not_the_original(self):
        event = DEFAULT_EVENT
        formatted_event = raw.format_event(event, 0, "_")
        assert event is not formatted_event

    def test_formatted_event_has_event_type(self):
        event_type = "__event_type__"
        event = raw.format_event(DEFAULT_EVENT, event_type, "_")
        assigned_type = event.get("__type")
        assert assigned_type is not None and assigned_type == event_type

    def test_formatted_event_has_timestamp(self):
        timestamp = get_timestamp()
        event = raw.format_event(DEFAULT_EVENT, "_", timestamp)
        assigned_timestamp = event.get("liverig__index__timestamp")
        assert assigned_timestamp is not None and assigned_timestamp == timestamp


class TestFormatAndSend:
    event_data = None

    def mock_connection_func(self, event):
        self.event_data = event

    def test_timestamp_prop_is_removed(self):
        raw.format_and_send({"timestamp": "_"}, "_", self.mock_connection_func)
        assert self.event_data.get("timestamp") is None

    def test_timestamp_is_formatted(self):
        raw.format_and_send({"timestamp": "_"}, "_", self.mock_connection_func)
        assert self.event_data.get("liverig__index__timestamp") is not None

    def test_timestamp_is_inserted(self):
        raw.format_and_send({}, "_", self.mock_connection_func)
        assert self.event_data.get("liverig__index__timestamp") is not None



# -*- coding: utf-8 -*-
from live_client.utils import logging, http

__all__ = ["fetch_resource"]


def fetch_resource(path, settings):
    live_settings = settings["live"]
    url = live_settings["url"]
    items_url = f"{url}{path}"
    items_list = []

    try:
        items_list = http.request_with_timeout(items_url, settings)
    except Exception:
        logging.exception(f"Error fetching {items_url}")

    return items_list



# -*- coding: utf-8 -*-
import urllib
import uuid

from live_client.utils import http
from live_client.utils.timestamp import get_timestamp
from live_client.events import annotation

__all__ = ["run_analysis", "analyse_and_annotate"]


def run_analysis(settings, **kwargs):
    live_settings = settings["live"]
    url = live_settings["url"]

    qs_data = {
        "assetId": kwargs.get("assetId"),
        "channel": kwargs.get("channel"),
        "qualifier": kwargs.get("channel"),
        "begin": kwargs.get("begin"),
        "end": kwargs.get("end"),
        "computeFields": kwargs.get(
            "computeFields", ["min", "max", "avg", "stdev", "linreg", "derivatives"]
        ),
    }

    params = urllib.parse.urlencode(qs_data, doseq=True)
    analysis_url = f"{url}/services/plugin-liverig-vis/auto-analysis/analyse?{params}"

    return http.request_with_timeout(analysis_url, settings)


def analyse_and_annotate(settings, **kwargs):
    analysis = run_analysis(settings, **kwargs)
    analysis.update(__src="auto-analysis", uid=str(uuid.uuid4()), createdAt=get_timestamp())
    return annotation.create(analysis, settings=settings)



# -*- coding: utf-8 -*-
import asyncio
import queue
from multiprocessing import Process, Queue

from eliot import start_action
from setproctitle import setproctitle
from aiocometd import Client

from live_client.events.constants import EVENT_TYPE_DESTROY, EVENT_TYPE_EVENT
from live_client.connection.rest_input import build_session
from live_client.utils.network import retry_on_failure, ensure_timeout
from live_client.utils import logging


__all__ = ["on_event", "run", "start", "watch"]


def start(statement, settings, timeout=None, **kwargs):
    live_settings = settings["live"]
    live_url = live_settings["url"]
    verify_ssl = live_settings.get("verify_ssl", True)

    if "session" not in settings:
        settings.update(session=build_session(live_settings))
    session = settings["session"]

    realtime = kwargs.get("realtime", False)
    span = kwargs.get("span", None)
    preload = kwargs.get("preload", False)
    max_retries = kwargs.get("max_retries", 0)

    api_url = f"{live_url}/rest/query"
    query_payload = [
        {
            "provider": "pipes",
            "preload": preload,
            "span": span,
            "follow": realtime,
            "expression": statement,
        }
    ]

    with retry_on_failure(timeout, max_retries=max_retries):
        logging.debug(f"Query '{statement}' started")
        r = session.post(api_url, json=query_payload, verify=verify_ssl)
        r.raise_for_status()

    channels = [item.get("channel") for item in r.json()]

    return channels


async def read_results(url, channels, output_queue):
    setproctitle("live-client: cometd client for channels {}".format(channels))

    with ensure_timeout(3.05):
        with start_action(action_type="query.read_results", url=url, channels=channels):
            # connect to the server
            async with Client(url) as client:
                for channel in channels:
                    logging.debug(f"Subscribing to '{channel}'")
                    await client.subscribe(channel)

                # listen for incoming messages
                async for message in client:
                    logging.debug(f"New message'{message}'")
                    output_queue.put(message)

                    # Exit after the query has stopped
                    event_data = message.get("data", {})
                    event_type = event_data.get("type")
                    if event_type == EVENT_TYPE_DESTROY:
                        return


def watch(url, channels, output_queue):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(read_results(url, channels, output_queue))


def run(statement, settings, timeout=None, **kwargs):
    with start_action(action_type="query.run", statement=statement):
        live_settings = settings["live"]

        channels = start(statement, settings, timeout=timeout, **kwargs)

        logging.debug(f"Results channel is {channels}")

        live_url = live_settings["url"]
        results_url = f"{live_url}/cometd"

        events_queue = Queue()
        process = Process(target=watch, args=(results_url, channels, events_queue))
        process.start()

    return process, events_queue


def on_event(statement, settings, realtime=True, timeout=None, **query_args):
    def handler_decorator(f):
        def wrapper(*args, **kwargs):
            results_process, results_queue = run(
                statement, settings, realtime=realtime, timeout=timeout, **query_args
            )
            last_result = None

            while True:
                try:
                    event = results_queue.get(timeout=timeout)
                except queue.Empty:
                    logging.exception(f"No results after {timeout} seconds")
                    break

                event_type = event.get("data", {}).get("type")
                if event_type == EVENT_TYPE_DESTROY:
                    break
                elif event_type != EVENT_TYPE_EVENT:
                    continue

                last_result = f(event, *args, **kwargs)

            results_process.join()

            return last_result

        return wrapper

    return handler_decorator



# -*- coding: utf-8 -*-
from live_client.utils.timestamp import get_timestamp
from live_client.utils import logging
from live_client.connection import autodetect

__all__ = ["create"]


def create(event_type, event_data, settings):
    connection_func = autodetect.build_sender_function(settings["live"])
    logging.debug(f'Creating raw event of type "{event_type}": {event_data}')
    format_and_send(event_data, event_type, connection_func=connection_func)


def format_and_send(event_data, event_type, connection_func):
    timestamp = event_data.pop("timestamp", get_timestamp())
    event = format_event(event_data, event_type, timestamp)
    connection_func(event)


def format_event(event_data, event_type, timestamp):
    event_data = event_data.copy()
    event_data["__type"] = event_type
    event_data["liverig__index__timestamp"] = timestamp
    return event_data



import os
from setuptools import setup, find_packages

__all__ = ["setup"]


def read_file(filename):
    """Read a file into a string"""
    path = os.path.abspath(os.path.dirname(__file__))
    filepath = os.path.join(path, filename)
    try:
        return open(filepath).read()
    except IOError:
        return ""


def get_readme():
    """Return the README file contents. Supports text,rst, and markdown"""
    for name in ("README", "README.rst", "README.md"):
        if os.path.exists(name):
            return read_file(name)
    return ""


def install_requires():
    requirements = read_file("requirements.txt")
    return requirements


setup(
    name="live_client",
    version="0.5.2",
    description="Client libraries to connect with the Intelie LIVE platform",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    scripts=["live_client/scripts/check_live_features"],
    url="https://github.com/intelie/liveclient-python",
    author="Vitor Mazzi",
    author_email="vitor.mazzi@intelie.com.br",
    license="Apache",
    install_requires=install_requires(),
    zip_safe=True,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Monitoring",
    ],
)



# -*- coding: utf-8 -*-
from live_client.resources.base import fetch_resource
from .team import fetch_teams_info

__all__ = ["fetch_user_info"]


def fetch_user_info(settings, include_teams=False):
    user = fetch_resource("/rest/me", settings)

    if user is not None:
        user_teams = user.get("teams", [])

        if include_teams and user_teams:
            team_ids = [item["id"] for item in user_teams]
            user_teams = fetch_teams_info(team_ids, settings)
            user.update(teams=user_teams)

    return user



# -*- coding:utf-8 -*-

ASSETS = "ASSETS"
MANAGE_WELLS = "MANAGE_WELLS"
DASHBOARDS = "DASHBOARDS"
USE_MESSENGER = "USE_MESSENGER"
MANAGE_DASHBOARDS = "MANAGE_DASHBOARDS"
MANAGE_ASSETS = "MANAGE_ASSETS"
EDIT_ANNOTATIONS = "EDIT_ANNOTATIONS"
MANAGE_PIPES_WIDGETS = "MANAGE_PIPES_WIDGETS"
VIEW_WELLS = "VIEW_WELLS"
MANAGE_MONITORING = "MANAGE_MONITORING"
CONSOLE = "CONSOLE"
CREATE_ROOMS = "CREATE_ROOMS"
IMPORT_DATA = "IMPORT_DATA"
ADMIN = "ADMIN"
ADVANCED_IMPORT_DATA = "ADVANCED_IMPORT_DATA"



# -*- coding: utf-8 -*-
from collections import UserDict
from datetime import datetime
from pytz import timezone

__all__ = ["Message"]


class Message(UserDict):
    def __init__(self, message_event):
        author_timezone = timezone(message_event.get("timeZone", "UTC"))
        created_at = author_timezone.localize(
            datetime.fromtimestamp(message_event.get("createdAt") / 1000)
        )

        message_event.update(text=message_event.get("message", ""), created_at=created_at)

        super().__init__(message_event)

    def has_mention(self, name):
        return name in self.data.get("text")

    def remove_mentions(self, name):
        self.data["text"] = self.data["text"].replace(name, "")
        return self

    def serialize(self):
        return self.data



# -*- coding: utf-8 -*-

from live_client.utils import http

__all__ = ["fetch_curves", "only_enabled_curves"]


def fetch_curves(settings):
    live_settings = settings["live"]
    host = live_settings["host"]

    url = "{}/services/plugin-liverig/curves/".format(host)

    data = http.request_with_timeout(url, settings)
    return data


def only_enabled_curves(curves):
    idle_curve = [{}]

    return dict(
        (name, value)
        for (name, value) in curves.items()
        if value.get("options", idle_curve) != idle_curve
    )
